/*
	PROforma, PROgs FOnt & Raster MAnager

	FastPattern drawing routines for mode 4 bitmap driver

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 9 November 1999
	21.10.2000 fixed small problem (only 7 of 8 bits in mask used)
*/

#include "PFmodule.h"

static const unsigned char bitmasks[]={ 0x80, 0x40, 0x20, 0x10, 8, 4, 2, 1};

Error FastPatternDrawMask(Bitmap bitmap, Mask *mask, short x, short y)
{
	short i,j;						/* loop counter */
	unsigned char *rlinestart;		/* curr pos in raster */
	unsigned short *rbase;			/* curr pos in bitmap */
	unsigned char *mbase, *mlinestart;	/* curr pos in mask */
	short linecount;				/* number of bytes to check in mask */
	unsigned char data; 			/* current part of char */
	int nrlines;
	int first=0, last, count;		/* first & last bit to display */
	int colour=*((unsigned short *)bitmap->pattern);

	mlinestart=mask->base;
	linecount=mask->linc;
	last=mask->xsiz;

	/* horizontal clipping */
	if (x<0 || mask->xsiz+x>bitmap->clip.xsiz)
	{
		/* still ok if completely invisible */
		if (x+mask->xsiz<=0 || x>=bitmap->clip.xsiz)
			return ERR_OK;
		else
		{
			/* the actual clipping */
			if (x<0)
			{
				first=(-x)&7;
				x=(-x)/8;
				mlinestart+=x;
				linecount-=x;
				last-=x*8;
				x=0;
			}
			if (last+x>bitmap->clip.xsiz)
				last=bitmap->clip.xsiz-x;
		}
	}

	/* vertical clipping */
	nrlines=mask->ysiz;
	if (y<0) { nrlines+=y; mlinestart+=(-y)*mask->linc; y=0; }
	if (nrlines<=0) return ERR_OK;
	if (nrlines+y > bitmap->clip.ysiz) nrlines-=nrlines+y-bitmap->clip.ysiz;
	if (nrlines<=0) return ERR_OK;

	x+=bitmap->clip.xorg;
	y+=bitmap->clip.yorg;

	rlinestart=bitmap->base+y*bitmap->linc+x*2;

	for( ; nrlines ; nrlines--)
	{
			rbase=(unsigned short *)rlinestart;
			mbase=mlinestart;
			count=0;
			for (i=linecount ; i ; i--)
			{
				unsigned char *bits=bitmasks;
				for (j=8 ; j && count<=last ; j--)
				{
					if ((*mbase & *bits++) && count>=first)
						*rbase=colour;
					rbase++;
					count++;
				}
				*mbase++;
			}
			rlinestart+=bitmap->linc;
			mlinestart+=mask->linc;
	}
	return ERR_OK;
}
