/*
	PROforma, PROGS Font & Raster Manager

	PFPageShow
		display the current page. If the page will be rendered in multiple
		passes, then the PageBbox will be adjusted to render the next pass. The
		currently finished pass will be sent to the device.
		If the last pass has been sent to the printer, the page will be ejected
		and the PageBbox will be adjusted to the first pass of a page.

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 17 September 1993
	08.11.1995 code still cleared the page for multi-pass pages - removed to be
			   consistent with documentation (the page will otherwise just be
			   cleared twice).
	24.10.1996 when PageOrigin was set explicitly, this was reversed when a
			   second page of a multi-pass page is rendered, and a
			   PFPageBboxRestore is now also performed for single pass buffers
*/

#include "gstate.h"
#include "system.h"

Error PFPageShow(Gstate gstate)
{
	Error err=ERR_OK;

	/* was this the first pass -> init */
	if (MEMSame(sizeof(PPageBbox),&gstate->pagebbox,&gstate->firstbbox))
		err=gstate->printer->InitPage(gstate);
	if (err) return err;

	/* show this pass */
	err=gstate->printer->ShowPage(gstate);
	if (err) return err;

	/* adjust PageBbox for next pass */
	if (gstate->passes!=1)
	{
		gstate->pagebbox.yporg+=short2pt(gstate->pagebbox.ysiz);
		gstate->Origin.y+=short2pt(gstate->pagebbox.ysiz);
		if (gstate->pagebbox.yporg+short2pt(gstate->pagebbox.ysiz) >
			gstate->allbbox.yporg+short2pt(gstate->allbbox.ysiz))
				gstate->pagebbox.ysiz=pt2short(
					gstate->allbbox.yporg+short2pt(gstate->allbbox.ysiz)-
					gstate->pagebbox.yporg);
		if (gstate->pagebbox.ysiz==0)
		{
			gstate->pagebbox=gstate->firstbbox;
			gstate->Origin.y-=gstate->passes*short2pt(gstate->pagebbox.ysiz);
		}
	}
	PFPageBboxRestore(gstate);

	/* was this the last pass -> stop */
	if (MEMSame(sizeof(PPageBbox),&gstate->pagebbox,&gstate->firstbbox))
		err=gstate->printer->StopPage(gstate);

	return err;
}

Error PFCopies(Gstate gstate, int count)
{
	return gstate->printer->SetCopies(gstate, count);
}
