/*
	PROforma, PROgs FOnt & Raster MAnager

	Routines to modify the PageBbox
	PFPageBboxReset
	PFPageBboxRestore
	PFPageOriginSet
	PFPageBboxSet
	PFPageBboxGet
	PFPageBboxFullGet

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 July 1993
	09.10.1995 changed naming scheme
	20.10.1995 fixed PageBboxRestore - was broken with introduction of Bitmap
	20.03.1996 added PFPPageOriginSet, PFPPageBboxSet, PFPPageBboxGet,
			   PFPPageBboxFullGet
	12.01.1997 PFPageBboxSet exchanged ysiz and xorg (?)
			   PF(P)PageBboxSet adjusted PageOrigin one pixel too little
	16.01.1997 PF(P)PageBboxSet adjust PageOrigin too much when PageBbox has
			   to be clipped. Bug introduced on 12.01
*/

#include "gstate.h"
#include "system.h"

Error PFPageBboxReset(Gstate gstate)
{
	Bitmap raster=gstate->raster;
	raster->clip.xsiz=gstate->pagebbox.xsiz;
	raster->clip.ysiz=gstate->pagebbox.ysiz;
	raster->clip.xorg=gstate->pagebbox.xorg;
	raster->clip.yorg=gstate->pagebbox.yorg;
	gstate->Origin.x =gstate->pagebbox.xporg;
	gstate->Origin.y =gstate->pagebbox.yporg;
	return ERR_OK;
}

Error PFPageBboxRestore(Gstate gstate)
{
	Bitmap raster=gstate->raster;
	gstate->Origin.x+=short2pt(gstate->pagebbox.xorg-gstate->raster->clip.xorg);
	gstate->Origin.y+=short2pt(gstate->pagebbox.yorg-gstate->raster->clip.yorg);
	raster->clip.xsiz=gstate->pagebbox.xsiz;
	raster->clip.ysiz=gstate->pagebbox.ysiz;
	raster->clip.xorg=gstate->pagebbox.xorg;
	raster->clip.yorg=gstate->pagebbox.yorg;
	
	return ERR_OK;
}

Error PFPageOriginSet(Gstate gstate, pt xporg, pt yporg)
{
	/* adjust the page origin */
	gstate->Origin.x+=fixmul(xporg,gstate->todev.x);
	gstate->Origin.y+=fixmul(yporg,gstate->todev.y);
	return ERR_OK;
}

Error PFPPageOriginSet(Gstate gstate, short xporg, short yporg)
{
	/* adjust the page origin */
	gstate->Origin.x=short2pt(xporg);
	gstate->Origin.y=short2pt(yporg);
	return ERR_OK;
}

Error PFPageBboxSet(Gstate gstate, pt xsiz, pt ysiz, pt xorg, pt yorg)
{
	/* The PageBbox is adjusted, relative to the previous values */
	/* The given coordinates are in original user space, being 72points/inch */

	/* calculate the new values */
	xorg=fixmul(xorg,gstate->todev.x);
	yorg=fixmul(yorg,gstate->todev.y);
	xsiz=fixmul(xsiz,gstate->todev.x);
	ysiz=fixmul(ysiz,gstate->todev.y);

	/* we always need some kind of window !! */
	if (xsiz<=pt_half || ysiz<=pt_half) { xsiz=0; ysiz=0; }

	/* adjust page origin */

	/* make coordinates absolute and integer */
	return PFPPageBboxSet(gstate, pt2rshort(xsiz), pt2rshort(ysiz),
								  pt2rshort(xorg), pt2rshort(yorg));
}

Error PFPPageBboxSet(Gstate gstate, short xsiz, short ysiz,
									short xorg, short yorg)
{
	PBbox *b=&gstate->raster->clip;

	/* The PageBbox is adjusted, relative to the previous values */
	/* The given coordinates are in pixels */

	/* try to make the PageBbox one pixel bigger in all directions, */
	/* this is expecially important for screen editing */
	xorg--; yorg--;
	xsiz+=2; ysiz+=2;

	/* make coordinates absolute */
	xorg+=b->xorg;
	yorg+=b->yorg;
	
	/* now test whether the given values are in range */
	/* and clip them if they are not */
	if (xorg < gstate->pagebbox.xorg)
	{
		short diff=gstate->pagebbox.xorg-xorg;
		xorg+=diff;
		xsiz-=diff;
	}
	if (yorg < gstate->pagebbox.yorg)
	{
		short diff=gstate->pagebbox.yorg-yorg;
		yorg+=diff;
		ysiz-=diff;
	}
	if (xsiz+xorg > gstate->pagebbox.xsiz+gstate->pagebbox.xorg)
		xsiz=gstate->pagebbox.xsiz+gstate->pagebbox.xorg-xorg;
	if (ysiz+yorg > gstate->pagebbox.ysiz+gstate->pagebbox.yorg)
		ysiz=gstate->pagebbox.ysiz+gstate->pagebbox.yorg-yorg;

	/* adjust PageOrigin */
	gstate->Origin.x+=short2pt(xorg-b->xorg);
	gstate->Origin.y+=short2pt(yorg-b->yorg);

	/* now set the new values */
	b->xsiz=xsiz;
	b->ysiz=ysiz;
	b->xorg=xorg;
	b->yorg=yorg;

	return ERR_OK;
}

Error PFPageBboxGet(Gstate gstate, PageBbox *result)
{
	PBbox *b=&gstate->raster->clip;
	PPageBbox *d=&gstate->pagebbox;

	result->xsiz = b->xsiz				 *gstate->ppp.x;
	result->ysiz = b->ysiz				 *gstate->ppp.y;
	result->xorg =(b->xorg-d->xorg) 	 *gstate->ppp.x;
	result->yorg =(b->yorg-d->yorg) 	 *gstate->ppp.y;
	result->xporg=fixmul(gstate->Origin.x,gstate->ppp.x);
	result->yporg=fixmul(gstate->Origin.y,gstate->ppp.y);
	return ERR_OK;
}

Error PFPageBboxFullGet(Gstate gstate, Bbox *result)
{
	PPageBbox *d=&gstate->allbbox;

	result->xsiz=		d->xsiz *gstate->ppp.x;
	result->ysiz=		d->ysiz *gstate->ppp.y;
	result->xorg=fixmul(d->xporg,gstate->ppp.x);
	result->yorg=fixmul(d->yporg,gstate->ppp.y);
	return ERR_OK;
}

Error PFPPageBboxGet(Gstate gstate, PPageBbox *result)
{
	PBbox *b=&gstate->raster->clip;
	PPageBbox *d=&gstate->pagebbox;

	result->xsiz =b->xsiz;
	result->ysiz =b->ysiz;
	result->xorg =b->xorg-d->xorg;
	result->yorg =b->yorg-d->yorg;
	result->xporg=gstate->Origin.x;
	result->yporg=gstate->Origin.y;
	return ERR_OK;
}

Error PFPPageBboxFullGet(Gstate gstate, PBbox *result)
{
	PPageBbox *d=&gstate->allbbox;

	result->xsiz=d->xsiz;
	result->ysiz=d->ysiz;
	result->xorg=d->xporg;
	result->yorg=d->yporg;
	return ERR_OK;
}
