/*
	PROforma, PROGS Font & Raster Manager

	routines to control the QVME screen driver - mode 4 only

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 28 September 1995
	02.11.1995 made alloc/realease window much safer
	14.12.1995 automatic recognition of mono/mode4
	21.06.1996 support for setting DisplayMode added
	27.11.1996 Handle used to fill in res for dummy driver
	16.01.1997 - driver now only one file
			   - PFPageShow now only updates area inside PageBbox and WindowSub
	22.12.1997 - when something passed as device string, then the driver will
				 draw in a save area, but not own a window !
	23.12.1997 - added support for PFPWindowLink call
			   - no more errors when window not valid or not open
	17.11.1999 added support for mode 33 on the Q40 (16bit colour)
	23.06.2000 added support for mode 32 (16bit QPC/QXL) MK (Marcel Kilgus)
*/

#include <mem.h>
#include <job.h>
#include <win.h>

#include "PFmodule.h"

static short screenXsize;	/* total screen size - horizontal */
static short screenYsize;	/* total screen size - vertical */
static WindowMode mode; 	/* screen mode */

extern DRIVdriver init; 	/* the actual driver description */

Error ColourMonoAdjust(Gstate gstate, pt *colour);
Error ColourMonoNearest(Gstate gstate, pt *colour, int *result);
Error ColourMonoDelta(Gstate gstate, pt *colour, pt *delta);
Error ColourMode32Adjust(Gstate gstate, ColourRGB *colour);
Error ColourMode32Nearest(Gstate gstate, ColourRGB *colour, int *result);
Error ColourMode32Delta(Gstate gstate, ColourRGB *colour, ColourRGB *delta);
Error ColourMode33Adjust(Gstate gstate, ColourRGB *colour);
Error ColourMode33Nearest(Gstate gstate, ColourRGB *colour, int *result);
Error ColourMode33Delta(Gstate gstate, ColourRGB *colour, ColourRGB *delta);

typedef struct _DriverInfo {
	Window window;			/* the window which displays the output */
	WindowArea *area;		/* the bitmap which will be displayed in it */
	WindowInfo info;		/* the size and position of the window */
	WindowInfo rest;		/* the position in the window to display the area */
	int xorg, yorg; 		/* origin in buffer to display the area */
	WindowArea *save;		/* saved buffer */
	Job update; 		/* job id of update job if exists */
	int mywindow;			/* TRUE if window owned by QVME driver */
} DriverInfo;

Error scrsize()
{
	Error err;
	WindowInfo limits;
	Window win;

	err=WINOpen(2,1,0,0,&win);
	if (err) return err;
	err=WINLimits(win,&limits);
	if (err) return err;
	screenXsize=limits.xsiz;
	screenYsize=limits.ysiz;
	WINClose(win);
	return err;
}

Error InitDriver()
{
	Error err;
	Job job;

	/* need to determine the size of the window - safely */
	err=JOBCall(scrsize,"determine screen size",8,1024,&job,0);
	if (err) return err;
	JOBWait(job,FOREVER);

	/* read screen mode and adjust bitmap when mono */
	WINMode(&mode);
	if (mode==WINDOWMODE_2)
	{
		init.bitmap.PlaneCount=1;
		init.bitmap.PlaneGrouping=BITMAP_NO_PLANES;
		init.ColourSpace=PF_COLOURSPACE_GRAYSHADE;
		init.colour.Gray.ColourAdjust =ColourMonoAdjust;
		init.colour.Gray.ColourNearest=ColourMonoNearest;
		init.colour.Gray.ColourDelta  =ColourMonoDelta;
	}
	/* part added by MK, 23.06.2000 */
	if (mode==WINDOWMODE_16BITQPC)
	{
		init.bitmap.PlaneCount=1;
		init.bitmap.PlaneDepth=16;
		init.bitmap.ByteGrouping=BITMAP_NO_BYTEGROUPING;
		init.bitmap.PlaneGrouping=BITMAP_NO_PLANES;
		init.ColourSpace=PF_COLOURSPACE_RGB;
		init.colour.RGB.ColourAdjust =ColourMode32Adjust;
		init.colour.RGB.ColourNearest=ColourMode32Nearest;
		init.colour.RGB.ColourDelta  =ColourMode32Delta;
	}
	/* end of part added by MK */
	if (mode==WINDOWMODE_16BITQ40)
	{
		init.bitmap.PlaneCount=1;
		init.bitmap.PlaneDepth=16;
		init.bitmap.ByteGrouping=BITMAP_NO_BYTEGROUPING;
		init.bitmap.PlaneGrouping=BITMAP_NO_PLANES;
		init.ColourSpace=PF_COLOURSPACE_RGB;
		init.colour.RGB.ColourAdjust =ColourMode33Adjust;
		init.colour.RGB.ColourNearest=ColourMode33Nearest;
		init.colour.RGB.ColourDelta  =ColourMode33Delta;
	}

	return err;
}

Error WindowMove(Gstate gstate, int dx, int dy)
{
	DriverInfo *di=gstate->PrinterExtra;

	if (!di->mywindow) return ERR_OK;

	di->info.xpos+=dx;
	di->info.ypos+=dy;
	di->rest.xpos=di->info.xpos&1;

	WINRedefine(di->window,0,0,
				di->info.xsiz+di->rest.xpos*2+(di->info.xsiz&1), di->info.ysiz,
				di->info.xpos&-2,di->info.ypos);
	return ERR_OK;
}

Error BufferSize(Gstate gstate, int xsiz, int ysiz, Size *size)
{
	Error err;
	err=gstate->bitmap->BufferSize(xsiz,ysiz,size);
	*size+=offsetof(WindowArea,data);
	return err;
}

Error BufferUpdate(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;

	WINAreaSave(di->window, &di->rest, di->xorg,di->yorg, di->area);
	return ERR_OK;
}

Error BufferSave(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Size size=di->area->line_increment*di->area->ysiz+offsetof(WindowArea,data);

	if (!di->save)
		if (MEMAllocate(size,&di->save)) return ERR_IMEM;
	MEMCopy(size,di->area,di->save);
	return ERR_OK;
}

Error BufferRestore(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Size size=di->area->line_increment*di->area->ysiz+offsetof(WindowArea,data);

	if (!di->save) return ERR_ISYN;
	MEMCopy(size,di->save,di->area);
	return ERR_OK;
}

Error InitInit(Gstate gstate, char *device, PBbox *area)
{
	Error err;
	DriverInfo *di;

	err=MEMAllocate(sizeof(DriverInfo),&di);
	if (err) return err;
	di->window=WINDOW_NONE;
	di->mywindow=FALSE;

	if (device && *device)
	{
		di->info.xsiz=di->info.ysiz=32000;
		di->info.xpos=di->info.ypos=0;
	}
	else
	{
		err=WINOpen(2,1,0,0,&di->window);
		di->mywindow=TRUE;
		if (err) { MEMRelease(di); return err; };
		err=WINLimits(di->window,&di->info);
		if (err) { MEMRelease(di); return err; };
	}

	gstate->PrinterExtra=di;
	gstate->todev.x = short2pt(screenXsize) / 720;
	gstate->todev.y = short2pt(screenYsize) / 540;
	gstate->ppp.x = short2pt(720) / screenXsize;
	gstate->ppp.y = short2pt(540) / screenYsize;

	MEMCopy(sizeof(WindowInfo),&di->info,area);

	return err;
}

Error InitDone(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;

	/* have to build the header of the WindowArea */
	di->area=(WindowArea *)gstate->raster->base;
	gstate->raster->base+=offsetof(WindowArea,data);
	di->area->flag=WINDOWAREA_FLAG;
	di->area->xsiz=gstate->raster->all.xsiz;
	di->area->ysiz=gstate->raster->all.ysiz;
	di->area->line_increment=gstate->raster->linc;
	di->area->mode=mode;

	/* fill in the info stuff in the DriverInfo structure */
	di->info.xsiz=di->rest.xsiz=gstate->pagebbox.xsiz;
	di->info.ysiz=di->rest.ysiz=gstate->pagebbox.ysiz;
	di->info.xpos=				pt2short(gstate->pagebbox.xporg);
	di->info.ypos=				pt2short(gstate->pagebbox.yporg);

	/* and set the window */
	return WindowMove(gstate,0,0);
}

Error CleanUp(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;

	if (di)
	{
		if (di->save) MEMRelease(di->save);
		if (di->mywindow) WINClose(di->window);
		MEMRelease(gstate->PrinterExtra);
	}
	return ERR_OK;
}

Error ShowPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	PBbox *b=&gstate->raster->clip;
	WindowInfo info;

	if (di->update)
	{
		JOBZap(di->update,ERR_OK);
		di->update=NUL;
	}

	if (di->mywindow)
	{
		info.xsiz=b->xsiz;
		info.ysiz=b->ysiz;
		info.xpos=b->xorg+(di->info.xpos&1);
		info.ypos=b->yorg;

		WINAreaRestore(di->window, &info, b->xorg, b->yorg, di->area);
	}
	else
	{
		int xorg=b->xorg, yorg=b->yorg, tmp;

		/* need some clipping on the visible area ! */
		info.xsiz=b->xsiz;
		info.ysiz=b->ysiz;
		info.xpos=b->xorg-di->xorg;
		info.ypos=b->yorg-di->yorg;

		if ((tmp=di->xorg-xorg) > 0)
		{
			info.xsiz-=tmp;
			info.xpos+=tmp;
			xorg=di->xorg;
		}
		if ((tmp=di->yorg-yorg) > 0)
		{
			info.ysiz-=tmp;
			info.ypos+=tmp;
			yorg=di->yorg;
		}
		if ((tmp=info.xsiz+info.xpos-di->rest.xsiz) >0)
			info.xsiz-=tmp;
		if ((tmp=info.ysiz+info.ypos-di->rest.ysiz) >0)
			info.ysiz-=tmp;
		if (info.xsiz<=0 || info.ysiz<=0) return ERR_OK;

		/* adjust in window */
		info.xpos+=di->rest.xpos;
		info.ypos+=di->rest.ypos;

		WINAreaRestore(di->window, &info, xorg, yorg, di->area);
	}
	return ERR_OK;
}

static update(DriverInfo *di)
{
	while (TRUE)
	{
		JOBSuspend(MYSELF,5,NULL);
		WINAreaRestore(di->window, &di->rest, di->xorg,di->yorg, di->area);
	}
}

Error DisplayMode(Gstate gstate, int mode)
{
	DriverInfo *di=gstate->PrinterExtra;

	switch(mode)
	{
	case PF_DISPLAYMODE_NORMAL:
		if (di->update) JOBZap(di->update,ERR_OK);
		di->update=NUL;
		break;
	case PF_DISPLAYMODE_UPDATE:
		if (!di->update)
			return JOBCall(update,"display screen",1,2048,&di->update,1,di);
		break;
	default:
		return ERR_IPAR;
	}
	return ERR_OK;
}

Error ColourNearest(Gstate gstate, ColourRGB *colour, int *result)
{
	*result=0;						/* assume black */
	if (colour->green>pt_hundred/2)
		*result+=2; 			/* add some green */
	if (colour->red  >pt_hundred/2)
		*result+=1; 			/* add some red */
	return ERR_OK;
}

Error ColourDelta(Gstate gstate, ColourRGB *colour, ColourRGB *delta)
{
	MEMCopy(sizeof(ColourRGB),colour,delta);
	delta->blue=colour->blue;
	if (colour->green>pt_hundred/2)
		delta->green-=pt_hundred;
	if (colour->red  >pt_hundred/2)
		delta->red	-=pt_hundred;
	return ERR_OK;
}

Error ColourMonoAdjust(Gstate gstate, pt *colour)
{
	if (*colour>=pt_hundred)
		*colour=0;
	else
	if (*colour<=0)
		*colour=pt_hundred;
	else
		*colour=pt_hundred-*colour;
	return ERR_OK;
}

Error ColourMonoNearest(Gstate gstate, pt *colour, int *result)
{
	*result=0;						/* assume white */
	if (*colour>pt_hundred/2)
		*result=1;					/* make it black */
	return ERR_OK;
}

Error ColourMonoDelta(Gstate gstate, pt *colour, pt *delta)
{
	*delta=*colour;
	if (*colour>pt_hundred/2)
		*delta-=pt_hundred;
	return ERR_OK;
}

/* part added by MK, 23.06.2000 */

Error ColourMode32Nearest(Gstate gstate, ColourRGB *colour, int *result)
{
	int red, green, blue;
	red  =(colour->red	/100)>>11; if (red	>0x1f) red	=0x1f;
	green=(colour->green/100)>>11; if (green>0x1f) green=0x1f;
	blue =(colour->blue /100)>>11; if (blue >0x1f) blue =0x1f;
	*result=((green & 3) << 14) + (blue << 8) + (red << 3) + (green >> 2);
	return ERR_OK;
}

Error ColourMode32Delta(Gstate gstate, ColourRGB *colour, ColourRGB *delta)
{
	MEMClear(sizeof(ColourRGB),delta);
	return ERR_OK;
}

Error ColourMode32Adjust(Gstate gstate,ColourRGB *colour)
{ return ERR_OK; }

/* end of part added by MK */

Error ColourMode33Nearest(Gstate gstate, ColourRGB *colour, int *result)
{
	int red, green, blue, res=0;
	red  =(colour->red	/100)>>10; if (red	>0x3f) red	=0x3f;
	green=(colour->green/100)>>10; if (green>0x3f) green=0x3f;
	blue =(colour->blue /100)>>10; if (blue >0x3f) blue =0x3f;
	if ((red&1)+(green&1)+(blue&1)>1) res=1;
	red  &=0x3e;
	green&=0x3e;
	blue &=0x3e;
	*result=res + blue + (red<<5) + (green<<10);
	return ERR_OK;
}

Error ColourMode33Delta(Gstate gstate, ColourRGB *colour, ColourRGB *delta)
{
	MEMClear(sizeof(ColourRGB),delta);
	return ERR_OK;
}

Error ColourMode33Adjust(Gstate gstate,ColourRGB *colour)
{ return ERR_OK; }


Error Handle(int command, Gstate gstate, int params)
{
	DriverInfo *di=gstate->PrinterExtra;

	switch (command)
	{
	case 1:
		gstate->todev.x = short2pt(screenXsize) / 720;
		gstate->todev.y = short2pt(screenYsize) / 540;
		gstate->ppp.x = short2pt(720) / screenXsize;
		gstate->ppp.y = short2pt(540) / screenYsize;
		return ERR_OK;
	case 2:
		/* handler for PFPWindowLink call */
		{
			int *val=&params;
			if (di->mywindow) return ERR_ISYN;
			di->window=(Channel)*val++;
			di->rest.xsiz=(short)*val++;
			di->rest.ysiz=(short)*val++;
			di->rest.xpos=(short)*val++;
			di->rest.ypos=(short)*val++;
			di->xorg=(short)*val++;
			di->yorg=(short)*val++;
		}
		return ERR_OK;
	}
	return ERR_NIMP;
}

static Error nimp() { return ERR_NIMP; }
static Error nothing()	{ return ERR_OK; }

#define InitPage		(Error (*)(Gstate))nothing
#define StopPage		(Error (*)(Gstate))nothing
#define ColourAdjust	(Error (*)(Gstate,ColourRGB *))nothing
#define Define			(Error (*)(char *name, char *value))nimp
#define SetCopies		(Error (*)(Gstate, int count))nimp
#define GetFullPageBbox (Error (*)(Gstate, Bbox *box))nimp
#define DefaultDevice	(Error (*)(Gstate, char *device))nimp

DRIVdriver init={
	NULL, PF_SCREENDRIVER,
	"mode 4 screen driver", 	/* name */
	1,1,						/* xdpi, ydpi */
	{                           /* mode 4 bitmap */
		2, 1,					/* 2 planes, each 1 pixel deep */
		BITMAP_NO_BYTEGROUPING, /* no byte grouping */
		1,						/* altenative bytes in each plane */
	},

	InitDriver,
	Define,

	InitInit,
	BufferSize,
	InitDone,
	CleanUp,

	InitPage,
	ShowPage,
	StopPage,

	SetCopies,
	DisplayMode,
	WindowMove,

	BufferUpdate,
	BufferSave,
	BufferRestore,

	PF_COLOURSPACE_RGB,
	{
		ColourAdjust,
		ColourNearest,
		ColourDelta
	},

	GetFullPageBbox,
	DefaultDevice,

	Handle
};
