/*
	PROforma, PROGS FOnt & Raster MAnager

	This file has all the definitions which concern the PROforma native
	font file and the commands in it.

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	parts of this file come from the path_h header file in LINEdesign v1
*/

#include "io.h"

/*	a font file is actually a header, followed by a collection of blocks */
typedef struct {
	char flag;					/* flag, should be '*' */
	char type;					/* type of information in block */
	long len;					/* length of block, not including header */
								/* this value has to be even */
} blockheader;

/* for now we define four kinds of blocks */
/* the order of the blocks is not important */

/* font information */
#define font_information 1
typedef struct {
	char FullName[PF_MAXFONTNAME];		/* name of font */
	char FamilyName[PF_MAXFONTFAMILY];	/* name of fontfamily */
	char Weight[PF_MAXFONTWEIGHT];		/* font weight */
	char Version[PF_MAXFONTVERSION];	/* version of font */
	char Notice[PF_MAXFONTNOTICE];		/* some message from creator */
	short ItalicAngle;			/* italicangle of font */
	long UniqueID;				/* PostScript (tm) uniqueid of font */
	short UnderLinePosition;	/* suggested position for underline */
	short UnderLineThickness;	/* suggested thickness for underline */
	short PaintType;			/* method for drawing, 0=fill, 2=stroke */
	short StrokeWidth;			/* width for stroking character */
	pt FontMatrix[6];			/* default fontmatrix */
	short FontBbox[4];			/* max bounds for characters */
	short BlueValues[10];		/* BlueValues, 0x8000 if not present */
	short OtherBlues[10];		/* OtherBlues, 0x8000 if not present */
	short FamilyBlues[10];		/* FamilyBlues, 0x8000 if not present */
	short FamilyOtherBlues[10]; /* FamilyOtherBlues, 0x8000 if not present */
	pt BlueScale;				/* BlueScale */
	short BlueShift;			/* BlueShift */
	short BlueFuzz; 			/* BlueFuzz */
	short StdHW, StdVW; 		/* dominant stem sizes */
	short StemSnapH[12];		/* common horizontal stem width */
	short StemSnapV[12];		/* common vertical stem widths */
} font_info;

/* font subroutines */
/* font outlines (glyphs) */
#define font_subroutines 2
#define font_glyphs 3

typedef struct {
	short code; 			/* unicode for character, or number code for subr */
	char *base; 			/* pointer to routine, relative with start of */
							/* block (not the block header!!) */
} ref;

typedef struct {
	short count;			/* number of entries in list */
	ref list[1];			/* list with references, count entries */
							/* start of routines */
} font_routs;

/* kerning table */
#define font_kerntable 4
typedef struct {
	short leftcode; 		/* unicode of left character */
	short rightcode;		/* unicode of right character */
	short adjust;			/* adjustment for character width, assume 1000pt size */
} kernkey;

typedef struct {
	short count;			/* number of entries in list */
	kernkey list[1];		/* list with kernkeys, count entries */
} font_kern;

/* charwidth table */
#define font_widthtable 5
typedef struct {
	short unicode;		   /* unicode of character */
	short width;		   /* width, assume 1000pt size */
} widthkey;

#define PF_WIDTHARRAY_FIRST 0x20
#define PF_WIDTHARRAY_LAST	0xfe

typedef struct {
	short count;			/* number of entries in list */
							/* some of the widths are also stored in the */
							/* array below for extra fast access to the most */
							/* often used characters */
	short array[PF_WIDTHARRAY_LAST-PF_WIDTHARRAY_FIRST+1];
	widthkey list[1];		/* list with kernkeys, count entries */
} font_width;

#define fc_hstem 1
#define fc_vstem 3
#define fc_vmoveto 4
#define fc_rlineto 5
#define fc_hlineto 6
#define fc_vlineto 7
#define fc_rrcurveto 8
#define fc_closepath 9
#define fc_callsubr 10
#define fc_return 11
#define fc_escape 12
#define 	fec_dotsection 0
#define 	fec_vstem3 1
#define 	fec_hstem3 2
#define 	fec_seac 6
#define 	fec_sbw 7
#define 	fec_div 12
#define 	fec_callothersubr 16
#define 	fec_pop 17
#define 	fec_setcurrentpoint 33
#define fc_hsbw 13
#define fc_endchar 14
#define fc_rmoveto 21
#define fc_hmoveto 22
#define fc_vhcurveto 30
#define fc_hvcurveto 31

#define font_stack_size 24
#define font_subrout_max 10

/* byte values, minval<=c<=maxval => c-139 */
#define minval 32
#define maxval 246
/* positive word, minposword<=c<=maxposword => [(c-247)*256]+w+108 */
#define minposword 247
#define maxposword 250
/* negative word, minnegword<=c<=maxnegword => -[(c-251)*256]-w-108 */
#define minnegword 251
#define maxnegword 254
/* long word, followed by 4 bytes, big endian order */
#define longword 255

/* here come the definition of how a font is stored in memory */

typedef struct font_ {
	struct font_ *next; 		/* pointer to next font in list */
	Size length;				/* length of font file */
								/* actual fontdata, exact copy of font file */
	int usercount;				/* number of users for this font */
	struct _CachedFont *cache;	/* all font cache entries for this font */
	font_info *info;			/* info table of current font */
	font_routs *subrs;			/* subrs table of current font */
	font_routs *glyphs; 		/* glyph table of current font */
	font_kern *kern;			/* kern table of current font */
	font_width *widths; 		/* the data width the character widths */
								/* the font image is loaded from here on */
	char header[24];			/* indicates type of font file, for now this */
								/* should be "PROforma native font    ", no \0 */
	char version[2];			/* version, e.g. "01", no \0 */
	blockheader firstblock;
} font;

typedef struct _fontline {
	struct _fontline *next; 		/* next font if fontmap */
	char FullName[PF_MAXFONTNAME];	/* full name of font */
	char filename[IO_MAXFULLNAME];	/* name of file with font */
} fontline;

typedef struct fontrec_ {
	struct fontrec_ *next;			/* next entry in list */
	font *base; 					/* base of font */
	char FullName[PF_MAXFONTNAME];	/* name of font */
} fontrec;

/* HINTING */
/* the special definitions which have to do with hinting */

#define MAXHINTS 20 	/* maximum number of hints (a font hint needs 2!) */
/* note that some fonts need more, e.g. Akka : 21 hint pairs, */

typedef struct {
	pt org, real;		/* in device space original and adjusted coordinate */
	pt factor;			/* scale factor for previous zone */
} hint;

#define ZONE_TOP	1
#define ZONE_BOTTOM 2
typedef struct {
	pt start, end;		/* range used by zone, start<end */
	char type;			/* type of zone */
} zone;

typedef struct {
	CTM matrix; 						/* transformation matrix */
	char useXhints, useYhints;			/* use hints ?? */
	/* character level hints */
	char calculated;					/* TRUE when hint data sorted etc. */
	short XhintCount, YhintCount;		/* number of active hints */
	hint Xhints[MAXHINTS];				/* actual hint data */
	hint Yhints[MAXHINTS];
	/* font level hints */
	short XsnapCount,YsnapCount;		/* number of snapping values */
	pt XSnap[13], YSnap[13];			/* snapping values */
	pt BlueShift;						/* current value for BlueShift */
	pt BlueFuzz;						/* current value for BlueFuzz */
	char ForceBold; 					/* ForceBold */
	short BlueCount;					/* number of BlueValues */
	zone BlueValues[10];				/* current BlueValues */
} hintdata;
